#
# $Id: Filesystem.pm,v 1.2.2.1.58.1 2004/11/06 01:04:56 avanish Exp $
#ident "$Source: /project/unixvm-cvs/src/sol/cmd/sdscvt/lib/Filesystem.pm,v $"
#
# $Copyrights: Copyright (c) 2009 Symantec Corporation.
# All rights reserved.
#
# THIS SOFTWARE CONTAINS CONFIDENTIAL INFORMATION AND TRADE SECRETS OF
# SYMANTEC CORPORATION.  USE, DISCLOSURE OR REPRODUCTION IS PROHIBITED
# WITHOUT THE PRIOR EXPRESS WRITTEN PERMISSION OF SYMANTEC CORPORATION.
#
# The Licensed Software and Documentation are deemed to be commercial
# computer software as defined in FAR 12.212 and subject to restricted
# rights as defined in FAR Section 52.227-19 "Commercial Computer
# Software - Restricted Rights" and DFARS 227.7202, "Rights in
# Commercial Computer Software or Commercial Computer Software
# Documentation", as applicable, and any successor regulations. Any use,
# modification, reproduction release, performance, display or disclosure
# of the Licensed Software and Documentation by the U.S. Government
# shall be solely in accordance with the terms of this Agreement.  $
#


use Disk;
use MD; use VxVM;
package Filesystem;
use Egettxt;
use OrElse;

use vars qw( @fs $rootdev $swapdev $rootencap $force_reboot $dumpdev_is_vol );

$rootdev = undef;
$swapdev = undef;

$rootencap = 0;
$force_reboot = 0;
$dumpdev_is_vol = 0;

@fs = ();

my @required_boot_fs = qw( / /usr /usr/platform );
my @force_reboot_fs = (@required_boot_fs, qw( /var /var/adm /usr/kvm ));
my @vxencap_rootable_fs = qw( /usr /var /opt );

sub SDSUnencap {
	my @res = ();

	# Make sure primary swap and required filesystems into plain
	# partitions for the next boot.
	#
	foreach my $fs (@fs) {
		#
		# Certain filesystems and all swap partitions require a
		# reboot after the conversion to properly use the newly
		# encapsulated volumes.
		#
		next unless ($fs->{'fstyp'} eq 'swap' or
		    (grep { $_ eq $fs->{'mountpoint'} } @force_reboot_fs));

		$force_reboot = 1;

		#
		# A subset of these must be available during the conversion
		# itself, so doconvert will change vfstab to use raw
		# partitions during the conversion.  We must preserve those
		# partitions and tell doconvert what to change in vfstab.
		#
		# NB: all swap would be in use, so to avoid forcing all of
		# those partitions to have simple partition representations,
		# doconvert will comment them out during the conversion.
		#
		next unless ($fs->{'olddev'} eq $swapdev or
		    (grep { $_ eq $fs->{'mountpoint'} } @required_boot_fs));

		#
		# If primary swap is a volume then we need to touch
		# /etc/vx/.dumpadm which will cause startup scripts to
		# trick the OS into dumping onto the same area of the
		# disk that the volume covers.
		#
		$dumpdev_is_vol = 1 if ($fs->{'olddev'} eq $swapdev);

		if ($VxVM::diskgroup ne 'rootdg') {
			Terminal('vmsolsds:92',
				'$prog: FATAL: Partition $part is required at boot but is being converted into $dg instead of rootdg',
				{ 'part' => $fs->{'olddev'},
				  'dg' => $VxVM::diskgroup } );
		}

		my ($part, @parts);

		if ((ref $fs) eq 'Filesystem::Encapsulate') {
			# This is already a partition in the vfstab so
			# it doesn't need any work at the doconvert step.
			# However, during S61convert we must not destroy
			# this partition (which will be mounted or used
			# for swap) so we mark it as protected.

			my ($name, $path) = Disk::NameAndPath($fs->{'olddev'});
			$part = Partition->Find($name);
			if ($part) {
				$part->Protect;
				@parts = ($part);
			} else {
				Terminal('vmsolsds:93',
					'$prog: FATAL: Filesystem $mtpt on partition $part cannot be encapsulated because the partition was not found in the VTOC',
					{ 'mtpt' => $fs->{'mountpoint'},
					  'part' => $fs->{'olddev'} });
			}
		} elsif ((ref $fs) eq 'Filesystem::Metadevice') {
			# We need to unencapsulate this metadevice during
			# the conversion boot.  doconvert will have to
			# rewrite the metadevice to a plain partition
			# (which must already exist as a mirror of the
			# metadevice, we don't change it here).
			my $number = $fs->{'number'};
			my $md = MD->Find($number);
			die unless $md;
			@parts = $md->SimplePartition;
			unless (@parts) {
				Terminal('vmsolsds:94',
					'One mirror of $md must be a simple partition to mount as $mtpt during conversion boot',
					{ 'md' => $md->Name,
					  'mtpt' => $fs->{'mountpoint'} });
			}

			# Look for the rootiest disk.
			@parts = map $_->[0],
				sort { $a->[1] <=> $b->[1] }
				map [ $_, $_->Rootness ], @parts;
			$part = $parts[0];
			push @res, "vfstabedit " . $fs->{'olddev'} . " " . $part->{'path'};
			$fs->{'tmpdev'} = $part->{'path'};

			#
			# During the conversion this partition must survive
			# (see above).  If it's the ROOT partition we preserve
			# them all in case they ever need to boot from an
			# alternate root.
			#
			if ($fs->{'mountpoint'} eq '/') {
				$_->Protect for @parts;
			} else {
				$part->Protect;
			}
		} else {
			die;
		}

		#
		# The clean attribute will bubble up to the plex level
		# so that we will resync the other mirrors after the
		# encapsulation (since we'll boot from the raw partition
		# and write on it with no mirroring)
		#
		$part->{'force_clean'} = 1;

		#
		# The converter needs to know the path(s) of the raw root
		# partition(s) to make appropriate changes to /etc/system.
		#
		if ($fs->{'mountpoint'} eq '/') {
			$rootencap = join("\n", map($_->{'path'}, @parts));
		}
	}
	(join "\n", @res) . "\n"
}

sub VFSTabEdits {
	(join "\n", map($_->VFSTabEdit, @fs)) . "\n"
}

sub estVxVM {
	my $nobj = 0;

	$nobj += $_->estVxVM for @fs;
	$nobj
}

sub toVxVM {
	map($_->toVxVM, @fs)
}

#
# Guess a use_type based on the mountpoint and fstyp
#
sub Vfstab2UseType {
	my ($mtpt, $fstyp) = @_;

	($mtpt =~ m,^/,) ?
		(($mtpt eq '/') ?  'root' : 'fsgen') :
		(($fstyp eq 'swap') ? 'swap' : 'gen')
}

sub ParseVFSTab {
	xopen VFSTAB, "< /etc/vfstab";
	filesystem: while (<VFSTAB>) {
		next if /^#/;

		my $saved = $_;
		my ($dev, $rdev, $mtpt, $fstyp, $rest) = split(/\s+/, $_, 5);
		next unless $dev =~ m,^/,;

		# root and primary swap
		$rootdev = $dev if ($mtpt eq '/');
		$swapdev = $dev if ($fstyp eq 'swap' and not $swapdev);

		if ($dev =~ m,^/dev/dsk/(c\d+t[A-F0-9]+d\d+s\d+), &&
		    (my $part = Partition->Find($1))) {
			# A filesystem on a partition which is part of a disk
			# we're going to convert.  The partition is going away
			# so we must encapsulate.
			$part->Use;
			push @fs, Filesystem::Encapsulate->new($part, $dev, $mtpt, $fstyp);
		} elsif ($dev =~ m,^/dev/md/(?:([^/]+)/)?dsk/d(\d+),) {
			my $diskset = $1 || '';
			my $number = $2;

			next filesystem if $diskset ne $MD::diskset;

			my $md = MD->Find($number);
			$md->{'force_use_type'} = Vfstab2UseType($mtpt, $fstyp);
			if (grep { $_ eq $mtpt } @vxencap_rootable_fs) {
				$_ = $mtpt;
				s,^/,,;
				$md->{'force_name'} = $_;
			}
			push @fs, Filesystem::Metadevice->new($number, $dev, $mtpt, $fstyp);
		}
	}
	close VFSTAB;
}

package Filesystem::Encapsulate;
use Egettxt;

sub estVxVM { 3 }

sub toVxVM {
	my $self = shift;

	my $part = $self->{'part'};
	my $name = $self->{'mountpoint'};
	$name = 'swapvol' if $name eq '-';
	$name = 'rootvol' if $name eq '/';

	$name =~ s=.*/(.*)=$1=;
	$name = VxVM->UniqueName($name, 2);

	my @sds = $part->toVxVM;

	my $plex = VxVM::Plex->new($name . "-p");
	$_->{'assoc'} = $plex for (@sds);
	$plex->{'aslist'} = \@sds;
	$plex->{'layout'} = 'concat';

	my $vol = VxVM::Volume->new($name);
	$plex->{'assoc'} = $vol;
	$vol->{'aslist'} = [ $plex ];
	$vol->{'use_type'} = Filesystem::Vfstab2UseType($self->{'mountpoint'},
							$self->{'fstyp'});
	$vol->{'comment'} = 'encap ' . $part->{'name'};
	$vol->{'cvtpath'} = $part->{'path'};

	$self->{'vol'} = $vol;
}

sub VFSTabEdit {
	my $self = shift;
	my $dev_at_convert = $self->{'tmpdev'} || $self->{'olddev'};

	$pathname = $self->{'vol'}->Path;
	if ($VxVM::diskgroup =~ /rootdg/) {
		my ( $path, $vol) = ($self->{'vol'}->Path =~ /(.*)\/([^\/]*$)/oi
);
		$pathname = $path . "/rootdg/" . $vol;
	}
	"vfstabedit " . $dev_at_convert . " " . $pathname;
}

sub new {
	my $type = shift;
	my ($part, $dev, $mtpt, $fstyp) = @_;

	bless { 'part' => $part,
		'olddev' => $dev,
		'mountpoint' => $mtpt,
		'fstyp' => $fstyp }, $type;
}

package Filesystem::Metadevice;
use Egettxt;

sub estVxVM { 0 }

sub toVxVM {
	my $self = shift;

	# These were already converted.
}

sub VFSTabEdit {
	my $self = shift;
	my $dev_at_convert = $self->{'tmpdev'} || $self->{'olddev'};


	my $md = MD->Find($self->{'number'});
	unless ($md) {
		Terminal('vmsolsds:95', '$prog: FATAL: Found mountpoint $mtpt in vfstab but did not find device $dev',
		{ 'mtpt' => $self->{'mountpoint'},
		  'dev' => $self->{'olddev'} });
		return;
	}

	my $pathname = $md->{'vol'}->Path;
	if ($VxVM::diskgroup =~ /rootdg/) {
		my ( $path, $vol) = ($md->{'vol'}->Path =~ /(.*)\/([^\/]*$)/oi);
		$pathname = $path . "/rootdg/" . $vol;
	}
	"vfstabedit " . $dev_at_convert . " " . $pathname;
}

sub new {
	my $type = shift;
	my ($number, $dev, $mtpt, $fstyp) = @_;

	bless { 'number' => $number,
		'olddev' => $dev,
		'mountpoint' => $mtpt,
		'fstyp' => $fstyp }, $type;
}

1;
