#
# $Id: VxVM.pm,v 1.2.2.1 2003/06/06 16:44:11 rohit Exp $
#ident "$Source: /project/unixvm-cvs/src/sol/cmd/sdscvt/lib/VxVM.pm,v $"
#
# $Copyrights: Copyright (c) 2009 Symantec Corporation.
# All rights reserved.
#
# THIS SOFTWARE CONTAINS CONFIDENTIAL INFORMATION AND TRADE SECRETS OF
# SYMANTEC CORPORATION.  USE, DISCLOSURE OR REPRODUCTION IS PROHIBITED
# WITHOUT THE PRIOR EXPRESS WRITTEN PERMISSION OF SYMANTEC CORPORATION.
#
# The Licensed Software and Documentation are deemed to be commercial
# computer software as defined in FAR 12.212 and subject to restricted
# rights as defined in FAR Section 52.227-19 "Commercial Computer
# Software - Restricted Rights" and DFARS 227.7202, "Rights in
# Commercial Computer Software or Commercial Computer Software
# Documentation", as applicable, and any successor regulations. Any use,
# modification, reproduction release, performance, display or disclosure
# of the Licensed Software and Documentation by the U.S. Government
# shall be solely in accordance with the terms of this Agreement.  $
#

use strict;

use Disk;
use Spacemap;
use Registered;

package VxVM;


use vars qw ( @ISA %VxVMTab $diskgroup );
@ISA = qw ( Registered );

#
# Name of the diskgroup being converted, set directly
#
$diskgroup = undef;

#
# All VxVM objects share a common registry indexed by Name,
# and provide a Rename alias to the Spacemap->Rekey method
#
%VxVMTab = ();
sub Registry { \%VxVMTab }
sub RegistryKey { 'name' }
sub RegistryReNew { undef }
sub Rename { my $self = shift; $self->Rekey(shift) }

#
# UniqueName(name[, extra]) makes a new object name based on the input,
# stripping illegal characters and adding numbers as necessary to get
# something unique, optionally leaving "extra" chars room for suffixes
# or prefixes (which are not guaranteed to be unique without calling again)
#
sub UniqueName {
	my $type = shift;
	my $name = shift;
	my $leftover = shift || 0;
	my $maxlen = 31 - $leftover;
	my $withnumlen = $maxlen - 3;

	# remove illegal chars
	$name =~ tr/A-Za-z0-9_-//cd;
	# chop to max length
	$name =~ s/(.{1,$maxlen}).*/$1/;
	my $n = '01';
	while ($type->Find($name)) {
		$name =~ s/(.{1,$withnumlen}).*/$1-$n/;
		$n = sprintf('%02d', $n + 1);
	}
	$name
}

#
# All VxVM objects define Vxmake, which outputs `vxmake -d' style descriptions
# of all component objects.
#
# Vxmake is called recursively (depth first) and the default action is to
# render the vxmake lines for all of the children of the object.
#
sub Vxmake {
	my $self = shift;

	map $_->Vxmake, @{$self->{'aslist'}};
}

package VxVM::Subdisk;

use vars qw ( @ISA );
@ISA = qw ( VxVM );

#
# Subdisks in vxmake description files use this notation to place them
# into columns.  We need this in order to properly place spacers.  If the
# parent is a concat they SHOULD all be in the same column and that is the
# default behavior of vxmake.
#
sub NameAndCol {
	my $self = shift;
	my $basename = $self->{'name'};

	if ($self->{'assoc'}->{'layout'} ne "concat") {
		$basename .= ":" . $self->{'column'} . "/" . $self->{'coloffs'};
	}
	$basename
}

sub Vxmake {
	my $self = shift;
	my $dm = $self->{'dm'};

	#
	# Might be a volume if we're layered:
	#
	my @rest = $self->{'dm'}->Vxmake;
	my @flags;

	push @rest, ("sd " . $self->{'name'} .  " dm_name=" . $dm->{'name'} .  " dm_offset=" . $self->{'offset'} .  " len=" . $self->{'len'} . " " . (join " ", @flags));
	#
	# Blank line after us if we're a subvolume, just to make the output
	# more human-readable.
	#
	push @rest, "" if ref $self->{'dm'} eq "VxVM::Volume";
	@rest
}

package VxVM::Plex;

use vars qw ( @ISA );
@ISA = qw ( VxVM );

sub Vxmake {
	my $self = shift;

	my @rest = $self->SUPER::Vxmake;

	my @flags = ();

	push @flags, "sd=" . join(",", map($_->NameAndCol, @{$self->{'aslist'}}));
	push @flags, "layout=" . $self->{'layout'};
	push @flags, "st_width=" . $self->{'st_width'} if $self->{'st_width'};
	push @flags, "len=" . $self->{'len'} if $self->{'len'};

	(@rest, "plex " . $self->{'name'} . " " . join(" ", @flags))
}

sub Assert {
	my $self = shift;

	# XXX NOTREACHED!
	die if $self->{'layout'} eq "stripe" and not defined $self->{'st_width'};
}

#
# Plexes synthesize the Clean attribute from their underlying subdisks,
# which receive it from their underlying partitions, which are clean if
# we plan to use the plain partition to boot from.  Kind of a misnomer, since
# plexes in volumes which have NO plexes reporting "clean" are started with
# `init active' which implies they're all clean (or at least of equal status).
#
# This test identifies cleanliness AFTER the conversion reboot and
# re-encapsulation by VxVM.
#
sub Clean {
	my $self = shift;

	# XXX deep search for layered if we needed that to work
	grep { $_->{'clean'} } @{$self->{'aslist'}};
}

#
# Fold unnecessary subvolumes.  This is only intended to handle cases
# that the MD.pm code creates.
#
sub Simplify {
	my $self = shift;

	# if this isn't part of a complete volume, don't get tricky!
	return unless $self->{'assoc'};

	#
	# concat plex containing exactly one subvolume:
	# remove the redundant pl-sv-v.  yes, that makes THIS plex redundant,
	# since the attributes we want to preserve are already on the
	# plex(es) under it
	#
	if ($self->{'layout'} eq 'concat' and @{$self->{'aslist'}} == 1 and
	    (ref $self->{'aslist'}->[0]->{'dm'}) eq 'VxVM::Volume') {
		my $vol = $self->{'assoc'};
		my $x_sv = $self->{'aslist'}->[0];
		my $x_vol = $x_sv->{'dm'};
		my $repl = $x_vol->{'aslist'};

		$self->Unregister;
		$x_sv->Unregister;
		$x_vol->Unregister;
		
		# if there's only one plex, rename it to be us
		if (@{$repl} == 1) {
			$repl->[0]->Rename($self->{'name'});
		}

		# replace this plex in parent aslist with child plexes
		@{$vol->{'aslist'}} = map { $_ == $self ? @{$repl} : $_ }
						@{$vol->{'aslist'}};

		# associate the underlying plexes with our parent
		$_->{'assoc'} = $vol for (@{$x_vol->{'aslist'}});

		# probably not necessary
		$x_vol->{'aslist'} = [];
		$self->{'assoc'} = undef;
		$self->{'aslist'} = [];
		$x_sv->{'assoc'} = undef;
	}
}


package VxVM::Volume;
use Egettxt;

use vars qw ( @ISA );
@ISA = qw ( VxVM );

sub Vxmake {
	my $self = shift;

	my @rest = $self->SUPER::Vxmake;

	my @flags = ();

	#
	# Layered volumes must share the use_type of the parent volume
	# or switchout/vxmake will split them up and be unable to create
	# them.
	#
	# XXX This is a problem for use_type=root, since there can be only
	# XXX one volume of use_type=root so the subvolumes CAN'T be in the
	# XXX same vxmake transaction (it dies with a cryptic error about
	# XXX objects not existing which are in the vxmake file).  Instead
	# XXX of working around this here, we just disassociate the problem
	# XXX mirrors and create them standalone.  They don't work very
	# XXX well on rootvol anyway, since vxconfigd can't autostart them.
	#
	if ($self->{'assoc'}) {
		push @flags, "layered=on";

		# Find superparent
		my $p = $self;
		$p = $p->{'assoc'} while ($p->{'assoc'});
		die unless ((ref $p) eq 'VxVM::Volume');
		push @flags, 'use_type=' . $p->{'use_type'};
	} else {
		push @flags, "use_type=" . $self->{'use_type'};
	}
	push @flags, 'plex=' . join(",", map($_->{'name'}, @{$self->{'aslist'}}));
	push @flags, 'len=' . $self->{'len'} if $self->{'len'};
	push @flags, 'read_pol=' . $self->{'read_pol'} if $self->{'read_pol'};
	push @flags, 'comment="' . $self->{'comment'} . '"' if $self->{'comment'};

	push @rest, ("vol " . $self->{'name'} . " " . join(" ", @flags));
	push @rest, "" unless $self->{'assoc'};
	@rest
}

sub Assert {
	# use_type set
	# use_type/read_pol match for raid
	# length set if assoc
}

sub Path {
	my $self = shift;

	die if $self->{'layered'};
	if ($VxVM::diskgroup eq 'rootdg') {
		return "/dev/vx/dsk/" . $self->{'name'}
	} else {
		return "/dev/vx/dsk/" . $VxVM::diskgroup . "/" . $self->{'name'}
	}
}

#
# Generate a `vxvol init' line for this volume that will make all the
# desired plexes clean (usually all of them).
#
sub VxvolInit {
	my $self = shift;

	my @clean = grep { $_->Clean } @{$self->{'aslist'}};

	my $cmd = $main::vxpath . "vxvol -g $VxVM::diskgroup init ";
	if (@clean) {
		$cmd .= "clean " . $self->{'name'} . " " .
					$clean[0]->{'name'} . "\n";
	} else {
		$cmd .= "active " . $self->{'name'} . "\n";
	}
	$cmd
}

#
# Remove any mirrors which are too complicated to be rootmirs.
# XXX need better criteria here, but no layered for sure, they won't start
# XXX properly and also the conversion vxmake would fail
#
# Don't worry about removing them all, since that would also mean there
# was no simple partition to boot off of, which is a Terminal error.
#
sub RemoveBadRootmirs {
	my $self = shift;
	my $newaslist = [];
	my $badaslist = [];

	for my $plex (@{$self->{'aslist'}}) {
		my $bad = 0;

		sd: for my $sd (@{$plex->{'aslist'}}) {
			if ((ref $sd->{'dm'}) eq 'VxVM::Volume') {
				$bad = 1;
				last sd;
			}
		}
		if ($bad) {
			Warning('vmsolsds:103',
				'$prog: Warning: Mirror $plex of $vol cannot be used as a root mirror under VxVM and will be converted but not attached to $vol',
				{ 'vol' => $self->{'name'},
				  'plex' => $plex->{'name'} });
			undef $plex->{'assoc'};
			push @{$badaslist}, $plex;
		} else {
			push @{$newaslist}, $plex;
		}
	}
	$self->{'aslist'} = $newaslist;

	#
	# vxmake is isn't very bright about layered volumes, so we need
	# any plexes we removed from rootvol to be associated to SOME
	# volume.  Otherwise it wants to make all the subdisks first,
	# even the ones that are really subvolumes, and that leads to
	# errors...
	#
	if (@{$badaslist}) {
		my $name = VxVM->UniqueName('rootvol-badmir');
		my $vol = VxVM::Volume->new($name);
		$_->{'assoc'} = $vol for (@{$badaslist});
		$vol->{'aslist'} = $badaslist;
		$vol->{'use_type'} = 'fsgen';
		$vol->{'len'} = $self->{'len'};
	}
}

package VxVM::DM;

use vars qw ( @ISA );
@ISA = qw ( VxVM );

#
# Create a private area if there are enough unused blocks
# at the start or end of the disk.  If that's not possible,
# make the disk nopriv.
#
sub AllocPriv {
	my $self = shift;
	my $disk = $self->{'disk'};
	my $privarea = $self->{'privarea'};

	return $privarea if $privarea;

	my $privneeded = $disk->RoundUp($main::privlen);
	my $spacemap = $disk->Spacemap();
	my $diskend = $disk->{'usableCylinders'} * $disk->{'secpercyl'};

	my @privlocs = ( [ 0, $privneeded -1 ],
			[ $diskend - $privneeded, $diskend - 1] );

	loc: foreach (@privlocs) {
		$privarea = $_, last loc unless $spacemap->Inuse(@{$_});
	}
	$spacemap->Add(@{$privarea}) if $privarea;
	$self->{'privarea'} = $privarea;
}

#
# Print the apropriate vxdisksetup line
#
sub Vxdisksetup {
	my $self = shift;
	my $disk = $self->{'disk'};
	my $privarea = $self->AllocPriv;

	if ($privarea) {
		my @args = ();
		push @args, "privoffset=" . $privarea->[0];
		push @args, "privlen=" . ($privarea->[1] - $privarea->[0] + 1);
		if ($privarea->[0] == 0) {
			# it should go here automatically, but beware
			# evil CDS muckery with our values XXX
			push @args, "puboffset=" . ($privarea->[1] + 1);
			$disk->{'devToPubOffset'} = -($privarea->[1] + 1);
		} else {
			push @args, "puboffset=0";
			# vxvm is going to skip the first block for us,
			# whether we like it or not.  setting puboffset=1
			# would round up to the next secpercyl, which we
			# don't want.
			$disk->{'devToPubOffset'} = -1;
		}
		my @protected = $disk->Protected;

		if (@protected) {
			@protected = map $_->{'part'}, @protected;
			push @args, "preserveslices=".(join ',', @protected);

			my %unused = ();
			$unused{$_} = 1 foreach (0 .. 7);
			$unused{2} = 0;
			$unused{$_} = 0 foreach (@protected);
			my @unused = sort { $b <=> $a } keys %unused;

			if (@unused < 2) {
				# XXX could switch to nopriv here, but that's
				# XXX not really desirable
				Terminal('vmsolsds:104',
'$prog: FATAL: Cannot allocate slices for priv/pub area on $name
$prog: Too many root mirrors being preserved?',
					{ 'name' => $self->{'name'} });
				# a bandaid to make it keep running
				@unused = (3, 4);
			}

			if (not $unused{3}) {
				my $privslice = shift @unused;
				$unused{$privslice} = 0;
				push @args, "privslice=$privslice";
			}
			if (not $unused{4}) {
				my $pubslice = shift @unused;
				$unused{$pubslice} = 0;
				push @args, "pubslice=$pubslice";
			}
		}
		# noreserve flag 
		push @args, "noreserve";
		return $main::vxdisksetup . " -i " . $self->{'name'} . " " .
				join(" ", @args). "\n";
	} else {
		# this seems to put things at the front of the disk
		# for real!  maybe we should spacer it anyway XXX
		$disk->{'devToPubOffset'} = 0;
		# XXX check 3.5
		return $main::vxpath . "vxdisk rm " . $self->{'name'} . "\n" .
			$main::vxpath . "vxdisk define " . $self->{'name'} . " type=nopriv\n";
	}
}

#
# this is invoked by ::Subdisk in case its DM is a Volume.  We don't
# need anything in Vxmake.  Not to be confused with Vxdisksetup!
#
sub Vxmake {
	()
}

1;
